@@ -1,12 +1,5 @@
-# This program is free software: you can redistribute it and/or modify it under the
-# terms of the Apache License (v2.0) as published by the Apache Software Foundation.
-#
-# This program is distributed in the hope that it will be useful, but WITHOUT ANY
-# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
-# PARTICULAR PURPOSE. See the Apache License for more details.
-#
-# You should have received a copy of the Apache License along with this program.
-# If not, see <https://www.apache.org/licenses/LICENSE-2.0>.
+# SPDX-FileCopyrightText: 2021 CmdKit Developers
+# SPDX-License-Identifier: Apache-2.0
 
 """
 Configuration management. Classes and interfaces for managing application level
@@ -17,7 +10,7 @@ local files and your environment.
 
 # type annotations
 from __future__ import annotations
-from typing import Tuple, IO, Dict, TypeVar, Callable, Union, Iterable, Optional, Any
+from typing import IO, Dict, TypeVar, Callable, Union, Iterable, Optional, Any
 
 # standard libs
 import os
@@ -36,64 +29,60 @@ DictItems: type = type({}.items())
 DictKeyIterator: type = type(iter({}))
 
 
+T = TypeVar('T')
+def _as_namespace(ns: T) -> Union[T, Namespace]:
+    """If `ns` is a mappable, coerce to Namespace, recursively, otherwise pass."""
+    return ns if not isinstance(ns, Mapping) else Namespace({k: _as_namespace(v) for k, v in dict(ns).items()})
+
+
+def _as_dict(ns: NSCoreMixin) -> dict:
+    """If `ns` is a mappable, coerce to dict, recursively, otherwise pass."""
+    return {k: v if not isinstance(v, Mapping) else _as_dict(v) for k, v in dict(ns).items()}
+
+
 class NSCoreMixin(dict):
     """Core namespace mechanics used by `Namespace` and `Configuration`."""
 
     def __init__(self, *args: Union[Iterable, Mapping], **kwargs: Any) -> None:
         """Initialize from same signature as `dict`."""
-        super().__init__(*args, **kwargs)
-
-    def __getitem__(self, key: str) -> Any:
-        """Like `dict.__getitem__` but return Namespace if value is a mappable."""
-        value = super().__getitem__(key)
-        if isinstance(value, Mapping):
-            return Namespace(value)
-        else:
-            return value
+        super().__init__()
+        for k, v in dict(*args, **kwargs).items():
+            self[k] = _as_namespace(v)
 
     def __setitem__(self, key: str, value: Any) -> None:
-        """Strip special type if `value` is Namespace."""
-        if isinstance(value, Mapping):
-            super().__setitem__(key, dict(value))
-        else:
+        """Strip special type if `value` is Namespace-like."""
+        if not isinstance(value, Mapping):
             super().__setitem__(key, value)
-
-    def get(self, key: Any, default: Any = None) -> Optional[Any]:
-        """Like `dict.get` but return Namespace if value is a mappable."""
-        value = super().get(key, default)
-        if isinstance(value, Mapping):
-            return Namespace(value)
         else:
-            return value
+            super().__setitem__(key, _as_namespace(dict(value)))
 
-    def items(self) -> Iterable[Tuple[str, Any]]:
-        """Yield key, value pairs."""
-        for key, value in super().items():
-            if isinstance(value, Mapping):
-                yield key, Namespace(value)
-            else:
-                yield key, value
+    def __setattr__(self, name: str, value: Any) -> None:
+        """Alias for index notation (if already present)."""
+        if name in self:
+            self[name] = value
+        else:
+            super().__setattr__(name, value)
 
     @classmethod
-    def _depth_first_update(cls, original: dict, new: dict) -> dict:
+    def __depth_first_update(cls, original: dict, new: dict) -> dict:
         """
-        Like normal `dict.update` but if values in both are mappable descend
+        Like normal `dict.update` but if values in both are mappable, descend
         a level deeper (recursive) and apply updates there instead.
         """
         for key, value in new.items():
             if isinstance(value, dict) and isinstance(original.get(key), dict):
-                original[key] = cls._depth_first_update(original.get(key, {}), value)
+                original[key] = cls.__depth_first_update(original.get(key, {}), value)
             else:
                 original[key] = value
         return original
 
     def update(self, *args, **kwargs) -> None:
-        """Implements a recursive, depth-first update (i.e., an "override")."""
-        self._depth_first_update(self, dict(*args, **kwargs))
+        """Depth-first update method."""
+        self.__depth_first_update(self, dict(*args, **kwargs))
 
     def __getattr__(self, item: str) -> Any:
         """
-        Get attribute by calling :meth:`__getitem__`.
+        Alias for index notation.
         Transparently expand `_env` and `_eval` variants.
         """
         variants = [f'{item}_env', f'{item}_eval']
@@ -101,16 +90,16 @@ class NSCoreMixin(dict):
             return self[item]
         for variant in variants:
             if variant in self:
-                return self._expand_attr(item)
+                return self.__expand_attr(item)
         else:
             raise AttributeError(f'missing \'{item}\'')
 
-    def _expand_attr(self, item: str) -> str:
+    def __expand_attr(self, item: str) -> str:
         """Interpolate values if `_env` or `_eval` present."""
 
         getters = {f'{item}': (lambda: self[item]),
-                   f'{item}_env': functools.partial(self._expand_attr_env, item),
-                   f'{item}_eval': functools.partial(self._expand_attr_eval, item)}
+                   f'{item}_env': functools.partial(self.__expand_attr_env, item),
+                   f'{item}_eval': functools.partial(self.__expand_attr_eval, item)}
 
         items = [key for key in self if key in getters]
         if len(items) == 0:
@@ -120,18 +109,17 @@ class NSCoreMixin(dict):
         else:
             raise ConfigurationError(f'\'{item}\' has more than one variant')
 
-    def _expand_attr_env(self, item: str) -> str:
+    def __expand_attr_env(self, item: str) -> str:
         """Expand `item` as an environment variable."""
         return os.getenv(str(self[f'{item}_env']), None)
 
-    def _expand_attr_eval(self, item: str) -> str:
+    def __expand_attr_eval(self, item: str) -> str:
         """Expand `item` as a shell expression."""
         return subprocess.check_output(str(self[f'{item}_eval']), shell=True).decode().strip()
 
     def __repr__(self) -> str:
         """Convert to string representation."""
-        original = super().__repr__()
-        return f'{self.__class__.__name__}({original})'
+        return f'{self.__class__.__name__}({repr(_as_dict(self))})'
 
 
 class Namespace(NSCoreMixin):
@@ -141,6 +129,7 @@ class Namespace(NSCoreMixin):
     Example:
         >>> ns = Namespace({'a': {'x': 1, 'y': 2}, 'b': 3})
         >>> ns.update({'a': {'x': 4, 'z': 5}})
+        >>> ns
         Namespace({'a': {'x': 4, 'y': 2, 'z': 5}, 'b': 3})
 
         >>> Namespace.from_local('config.toml', ignore_if_missing=True)
@@ -200,7 +189,7 @@ class Namespace(NSCoreMixin):
 
     def to_dict(self) -> Dict[str, Any]:
         """Explicitly coerce a Namespace to dictionary."""
-        return dict(self)
+        return _as_dict(self)
 
     def to_local(self, filepath: str, **options) -> None:
         """Output to local file. Format based on file extension."""
@@ -389,16 +378,16 @@ class Environ(NSCoreMixin):
         Collapse a namespace down to a single level by merging keys with their
         parent section by underscore.
 
-    Example:
-        >>> env = Namespace.from_env('MYAPP')
-        >>> env
-        Environ({'MYAPP_A_X': '1', 'MYAPP_A_Y': '2', 'MYAPP_B': '3'})
+        Example:
+            >>> env = Namespace.from_env('MYAPP')
+            >>> env
+            Environ({'MYAPP_A_X': '1', 'MYAPP_A_Y': '2', 'MYAPP_B': '3'})
 
-        >>> env.expand()
-        Environ({'a': {'x': 1, 'y': 2}, 'b': 3})
+            >>> env.expand()
+            Environ({'a': {'x': 1, 'y': 2}, 'b': 3})
 
-        >>> env.expand().flatten(prefix='MYAPP')
-        Environ({'MYAPP_A_X': '1', 'MYAPP_A_Y': '2', 'MYAPP_B': '3'})
+            >>> env.expand().flatten(prefix='MYAPP')
+            Environ({'MYAPP_A_X': '1', 'MYAPP_A_Y': '2', 'MYAPP_B': '3'})
         """
         ns = self.__class__(defaults=_flatten(self, prefix=prefix))
         ns._prefix = prefix
@@ -438,18 +427,21 @@ class Configuration(NSCoreMixin):
         1
     """
 
-    namespaces: Namespace = None
+    local: Namespace  # NOTE: used to track changes to the 'self'
+    namespaces: Namespace
 
     def __init__(self, **namespaces: Namespace) -> None:
         """Retain source `namespaces` and create master namespace."""
         super().__init__()
+        self.local = Namespace()
         self.namespaces = Namespace()
-        # self._master = Namespace()
         self.extend(**namespaces)
 
     def __repr__(self) -> str:
         """String representation of Configuration."""
         kwargs = ', '.join([f'{k}=' + repr(v) for k, v in self.namespaces.items()])
+        if self.local:
+            kwargs += f', _={repr(self.local)}'
         return f'{self.__class__.__name__}({kwargs})'
 
     def extend(self, **others: Union[Namespace, Environ]) -> None:
@@ -466,8 +458,11 @@ class Configuration(NSCoreMixin):
                           three=Namespace({'y': 5, 'u': {'i': 6, 'j': 7}})
         """
         for name, mapping in others.items():
-            self.namespaces[name] = Namespace(mapping)
-            self.update(self.namespaces[name])
+            if name != '_':
+                self.namespaces[name] = Namespace(mapping)
+                super().update(self.namespaces[name])
+            else:
+                self.local.update(mapping)
 
     @classmethod
     def from_local(cls, *, env: bool = False, prefix: str = None,
@@ -510,13 +505,56 @@ class Configuration(NSCoreMixin):
             >>> conf.which('u', 'i')
             'three'
         """
-        for label in reversed(list(self.namespaces.keys())):
+        namespaces = Namespace({**self.namespaces, '_': self.local})
+        for label in reversed(list(namespaces.keys())):
             try:
-                sub = self.namespaces[label]
+                sub = namespaces[label]
                 for p in path:
                     sub = sub[p]
                 return label
             except KeyError:
                 pass
         else:
-            raise KeyError(f'not found: {path}')
+            raise KeyError(f'Not found: {path}')
+
+    def __setattr__(self, name: str, value: Any) -> None:
+        """Intercept parameter assignment."""
+        if name in self:
+            self.update({name: value})
+        else:
+            super().__setattr__(name, value)
+
+    def update(self, *args, **kwargs) -> None:
+        """
+        Update current namespace directly.
+
+        Note:
+            The :class:`Configuration` class is itself a :class:`Namespace`-like object.
+            Doing any in-place changes to its underlying `self` does not change its member namespaces.
+            This may otherwise cause confusion about the provenance of those parameters.
+            Instead, overrides have been implemented to capture these changes in a `local` namespace.
+            If you ask :func:`which` namespace a parameter has come from and it was an in-place change,
+            it will be considered a member of the "_" namespace.
+
+        Example:
+            >>> conf = Configuration(a=Namespace(x=1))
+            >>> conf
+            Configuration(a=Namespace({'x': 1}))
+
+            >>> conf.update(y=2)
+            >>> conf
+            Configuration(a=Namespace({'x': 1}), _=Namespace({'y': 2}))
+
+            >>> conf.x = 2
+            >>> conf
+            Configuration(a=Namespace({'x': 1}), _=Namespace({'x': 2, 'y': 2}))
+
+            >>> conf.update(y=3)
+            >>> conf
+            Configuration(a=Namespace({'x': 1}), _=Namespace({'x': 2, 'y': 3}))
+
+            >>> dict(conf)
+            {'x': 2, 'y': 3}
+        """
+        self.local.update(*args, **kwargs)
+        super().update(*args, **kwargs)
