@@ -1,12 +1,5 @@
-# This program is free software: you can redistribute it and/or modify it under the
-# terms of the Apache License (v2.0) as published by the Apache Software Foundation.
-#
-# This program is distributed in the hope that it will be useful, but WITHOUT ANY
-# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
-# PARTICULAR PURPOSE. See the Apache License for more details.
-#
-# You should have received a copy of the Apache License along with this program.
-# If not, see <https://www.apache.org/licenses/LICENSE-2.0>.
+# SPDX-FileCopyrightText: 2021 CmdKit Developers
+# SPDX-License-Identifier: Apache-2.0
 
 """Unit tests for `cmdkit.config` behavior and interfaces."""
 
@@ -89,7 +82,7 @@ def test_namespace_update_simple() -> None:
 
 
 def test_namespace_update_complex() -> None:
-    """Test Namespace.update() behavior"""
+    """Test Namespace.update() behavior."""
 
     ns = Namespace({'a': {'x': 1, 'y': 2}})
 
@@ -103,6 +96,18 @@ def test_namespace_update_complex() -> None:
     assert ns == {'a': {'x': 5, 'y': 2, 'z': 3}, 'b': {'z': 3}}
 
 
+def test_namespace_deep_inplace_assignment() -> None:
+    """Test Namespace attribute access behavior."""
+
+    ns = Namespace({'a': {'x': 1, 'y': 2}})
+    assert isinstance(ns.a, Namespace)
+    assert ns.a == {'x': 1, 'y': 2}
+    assert ns.a.x == 1 and ns.a.y == 2
+
+    ns.a.x = 3
+    assert ns.a.x == 3
+
+
 TEST_DICT = {
     'a': {'a': 1, 'aa': 'aaa'},
     'b': {'b': 2, 'bb': 'bbb'},
@@ -265,9 +270,9 @@ def test_environ() -> None:
     """Test environment variable initialization along with Environ.reduce()."""
 
     # clean environment of any existing variables with the item
-    PREFIX = 'CMDKIT'
+    prefix = 'CMDKIT'
     for var in dict(os.environ):
-        if var.startswith(PREFIX):
+        if var.startswith(prefix):
             os.environ.pop(var)
 
     # populate environment with test variables
@@ -276,8 +281,8 @@ def test_environ() -> None:
         os.environ[field] = value
 
     # test base level Namespace|Environ equivalence
-    assert Namespace.from_env(prefix=PREFIX) == Environ(prefix=PREFIX)
-    assert Environ(prefix=PREFIX).expand() == Namespace(TEST_DICT)
+    assert Namespace.from_env(prefix=prefix) == Environ(prefix=prefix)
+    assert Environ(prefix=prefix).expand() == Namespace(TEST_DICT)
 
 
 TEST_ENV_TYPES = """\
@@ -294,9 +299,9 @@ def test_environ_expand() -> None:
     """Test automatic type coercion with Environ.reduce()."""
 
     # clean environment of any existing variables with the item
-    PREFIX = 'CMDKIT'
+    prefix = 'CMDKIT'
     for var in dict(os.environ):
-        if var.startswith(PREFIX):
+        if var.startswith(prefix):
             os.environ.pop(var)
 
     # populate environment with test variables
@@ -304,7 +309,7 @@ def test_environ_expand() -> None:
         field, value = line.strip().split('=')
         os.environ[field] = value
 
-    env = Environ(prefix=PREFIX).expand()
+    env = Environ(prefix=prefix).expand()
     assert isinstance(env['int'], int) and env['int'] == 1
     assert isinstance(env['float'], float) and env['float'] == 3.14
     assert isinstance(env['true'], bool) and env['true'] is True
@@ -317,9 +322,9 @@ def test_environ_flatten() -> None:
     """Test round-trip Environ('...').expand().flatten()."""
 
     # clean environment of any existing variables with the item
-    PREFIX = 'CMDKIT'
+    prefix = 'CMDKIT'
     for var in dict(os.environ):
-        if var.startswith(PREFIX):
+        if var.startswith(prefix):
             os.environ.pop(var)
 
     # populate environment with test variables
@@ -327,17 +332,17 @@ def test_environ_flatten() -> None:
         field, value = line.strip().split('=')
         os.environ[field] = value
 
-    env = Namespace.from_env(PREFIX)
-    assert env == env.expand().flatten(prefix=PREFIX)
+    env = Namespace.from_env(prefix)
+    assert env == env.expand().flatten(prefix=prefix)
 
 
 def test_environ_defaults() -> None:
     """Test defaults for missing environment variables."""
 
     # clean environment of any existing variables with the item
-    PREFIX = 'CMDKIT'
+    prefix = 'CMDKIT'
     for var in dict(os.environ):
-        if var.startswith(PREFIX):
+        if var.startswith(prefix):
             os.environ.pop(var)
 
     # populate environment with test variables
@@ -346,7 +351,7 @@ def test_environ_defaults() -> None:
         os.environ[field] = value
 
     # add default
-    env = Environ(prefix=PREFIX, defaults={'CMDKIT_DEFAULT_VALUE': '42'}).reduce()
+    env = Environ(prefix=prefix, defaults={'CMDKIT_DEFAULT_VALUE': '42'}).reduce()
     value = env['default']['value']
     assert isinstance(value, int) and value == 42
 
@@ -453,9 +458,9 @@ def test_configuration_from_local() -> None:
             output.write(data)
 
     # clean environment of any existing variables with the item
-    PREFIX = 'CMDKIT'
+    prefix = 'CMDKIT'
     for var in dict(os.environ):
-        if var.startswith(PREFIX):
+        if var.startswith(prefix):
             os.environ.pop(var)
 
     # populate environment with test variables
@@ -465,7 +470,7 @@ def test_configuration_from_local() -> None:
 
     # build configuration
     default = Namespace.from_toml(StringIO(TEST_CONFIG_DEFAULT))
-    cfg = Configuration.from_local(default=default, env=True, prefix=PREFIX,
+    cfg = Configuration.from_local(default=default, env=True, prefix=prefix,
                                    system=f'{TMPDIR}/system.toml',
                                    user=f'{TMPDIR}/user.toml',
                                    local=f'{TMPDIR}/local.toml')
@@ -475,7 +480,7 @@ def test_configuration_from_local() -> None:
     assert cfg.namespaces['system'] == Namespace.from_toml(StringIO(TEST_CONFIG_SYSTEM))
     assert cfg.namespaces['user'] == Namespace.from_toml(StringIO(TEST_CONFIG_USER))
     assert cfg.namespaces['local'] == Namespace.from_toml(StringIO(TEST_CONFIG_LOCAL))
-    assert cfg.namespaces['env'] == Environ(PREFIX).reduce()
+    assert cfg.namespaces['env'] == Environ(prefix).reduce()
 
     # verify parameter lineage
     assert cfg['a']['var0'] == 'default_var0' and cfg.which('a', 'var0') == 'default'
@@ -484,3 +489,27 @@ def test_configuration_from_local() -> None:
     assert cfg['b']['var3'] == 'local_var3' and cfg.which('b', 'var3') == 'local'
     assert cfg['c']['var4'] == 'env_var4' and cfg.which('c', 'var4') == 'env'
     assert cfg['c']['var5'] == 'env_var5' and cfg.which('c', 'var5') == 'env'
+
+
+def test_configuration_live_update() -> None:
+    """Do not allow the direct use of `update` on Configuration class."""
+
+    cfg = Configuration(a=Namespace(x=1))
+    assert repr(cfg) == 'Configuration(a=Namespace({\'x\': 1}))'
+    assert dict(cfg) == {'x': 1}
+    assert cfg.which('x') == 'a'
+
+    cfg.x = 2
+    assert repr(cfg) == 'Configuration(a=Namespace({\'x\': 1}), _=Namespace({\'x\': 2}))'
+    assert dict(cfg) == {'x': 2}
+    assert cfg.which('x') == '_'
+
+    cfg.update(y=2)
+    assert dict(cfg) == {'x': 2, 'y': 2}
+    assert repr(cfg) == 'Configuration(a=Namespace({\'x\': 1}), _=Namespace({\'x\': 2, \'y\': 2}))'
+    assert cfg.which('y') == '_'
+
+    cfg.update(y=3)
+    assert dict(cfg) == {'x': 2, 'y': 3}
+    assert repr(cfg) == 'Configuration(a=Namespace({\'x\': 1}), _=Namespace({\'x\': 2, \'y\': 3}))'
+    assert cfg.which('y') == '_'
